"""
Workflow for locking config
"""

from archimedes_config.workflows.workflow_base import WorkflowBase


class LockConfig(WorkflowBase):
    """workflow for locking secret"""

    def set_encrypted_key(self, encryption_key=None):
        """Sets encryption key"""
        if self.config_has_remote_conn_parameters():
            if encryption_key:
                raise Exception(
                    "Config set to connect to azure keyvault. `encryption_key` should not be set."
                )
            self.conf.set_default_key_from_key_vault()
        else:
            self.conf.set_default_key(encryption_key)

        if not self.conf.is_config_encrypted:
            self.conf.encrypt_configs()

    def _loop_for_secret(self):
        """
        Bypass parent class prompts to add secrets
        """

    def save(self, file_name, save_decrypted=False):
        """Saves the config"""
        super().save(file_name, save_decrypted=save_decrypted)



"""
Client for handling interactions with keyvault
"""

import os
from typing import Union

from azure.core.exceptions import ResourceNotFoundError  # pylint:disable=import-error
from azure.identity import (  # pylint:disable=import-error
    DefaultAzureCredential,
    EnvironmentCredential,
)
from azure.keyvault.secrets import SecretClient  # pylint:disable=import-error


class AzureKeyVaultClient:
    """
    Python client to handle communications with Azure keyvault
    """

    def __init__(self, vault_name):
        """
        Init method for KeyVaultClient
        :param vault_name: Name of the vault
        """
        self.credential = self.get_credential()
        self.vault_url = f"https://{vault_name}.vault.azure.net/"
        self.client = SecretClient(vault_url=self.vault_url, credential=self.credential)

    @staticmethod
    def get_credential():
        """
        Generates az credentials for authentication

        This method expects
            AZURE_TENANT_ID
            AZURE_CLIENT_ID
            AZURE_CLIENT_SECRET
        set as environments.

        If these are unavailable, az cli login will be used.

        :return:
        """
        if (
            "AZURE_TENANT_ID" in os.environ
            and "AZURE_CLIENT_ID" in os.environ
            and "AZURE_CLIENT_SECRET" in os.environ
        ):
            print("Using environment variables")
            credential = EnvironmentCredential()
        else:
            print("Using default creds")
            credential = DefaultAzureCredential()

        return credential

    def get_secret(
        self, name: str, return_none_if_not_found: bool = True
    ) -> Union[str, None]:
        """
        Retrieves secret for provided name from key vaults


        :param name: Key of secret to import
        :param return_none_if_not_found:
            if true and key not found, returns None
            if false and key not found, raise exception
        :return: Secret
        """

        try:
            secret_value = self.client.get_secret(name.replace("_", "-")).value
            return secret_value
        except ResourceNotFoundError as err:
            if "SecretNotFound" in err.message:
                if return_none_if_not_found:
                    return None
            raise err



"""
Utils related to path
"""

import os
from pathlib import Path


def find_repo_root() -> Path:
    """Finds root directory of repo"""

    current_path = Path(".").absolute().resolve()
    while len(current_path.parents) > 0:
        contents = os.listdir(current_path)
        if ".git" not in contents:
            current_path = current_path.resolve().parent
            continue
        return current_path
    raise Exception("Unable to detect the repo root.")


def resolve_filename_to_base(filename) -> str:
    """Resolved provided local filepath wrt project base"""
    root_path = find_repo_root()

    filename = Path(".").absolute() / filename
    if str(filename).startswith(str(root_path)):
        input_path = filename.relative_to(root_path)
        return str(input_path).replace("\\", "/")
    raise Exception("Config must be set within root path.")


def clean_filename(function):
    """Removes .toml extension out of filename"""

    def wrapper(*args, **kwargs):
        """Wrapper for cleaning file name"""
        if "file_name" in kwargs:
            if kwargs["file_name"] is not None and kwargs["file_name"].endswith(
                ".toml"
            ):
                kwargs["file_name"] = kwargs["file_name"][:-5]
        return function(*args, **kwargs)

    return wrapper



"""
Workflow to create a new secret
"""

from cryptography.fernet import Fernet

from archimedes_config.workflows.workflow_base import WorkflowBase


class ConfigCreator(WorkflowBase):
    """Class for creating secret"""

    def __init__(self):
        super().__init__()
        self.new_key_generated = True

    def load_file(self, file_name):
        """Creates a new config"""
        self.conf.create()

    def set_encrypted_key(self, encryption_key=None):
        """Creates a new encryption key"""
        while True:
            print("Do you want to generate a new encryption key")
            print("Or add a key from Az KeyVault?")
            option = input(
                "Enter `N` to generate OR `C` to register Az KeyVault."
            ).lower()
            if option == "n":
                self.conf.set_default_key(Fernet.generate_key().decode())
            elif option == "c":
                if not hasattr(self.conf, "set_default_key_from_key_vault"):
                    raise ImportError(
                        "Environment has not been configured to handle connections with azure key vault."
                    )
                self.new_key_generated = False
                vault_name = input("Enter vault name:")
                self.conf.add_new_config(
                    group_name="AZURE_KEYVAULT",
                    key_name="VAULT_NAME",
                    encrypted=False,
                    unencrypted_values=vault_name,
                    create_group_if_not_exist=True,
                    allow_updating=True,
                )
                vault_key = input("Enter vault key:")

                self.conf.add_new_config(
                    group_name="AZURE_KEYVAULT",
                    key_name="VAULT_KEY",
                    encrypted=False,
                    unencrypted_values=vault_key,
                    create_group_if_not_exist=True,
                    allow_updating=True,
                )
                self.conf.set_default_key_from_key_vault()
            else:
                print("Invalid input.")
                continue
            break



"""
Base module for workflows
"""
from archimedes_config.config_manager import ConfigManager
from archimedes_config.keyvault_config_manager import KeyvaultConfigManager


class WorkflowBase:
    """Base class for workflows"""

    def __init__(self):
        self.group_name = None
        try:
            self.conf = KeyvaultConfigManager()
        except ImportError:
            self.conf = ConfigManager()

    def config_has_remote_conn_parameters(self):
        """
        Checks if config can accesss reote key vault
        """
        try:
            self.conf.get("AZURE_KEYVAULT", "VAULT_NAME")
            self.conf.get("AZURE_KEYVAULT", "VAULT_KEY")
        except KeyError:
            return False
        if hasattr(self.conf, "set_default_key_from_key_vault"):
            return True
        return False

    def _get_encrypted(self):
        """Gets encryption status from user"""
        encrypted = input("Do you want the config to be encrypted?[Y/N]:").lower()
        if encrypted == "y":
            return True
        if encrypted == "n":
            return False
        print("Invalid input!")
        return self._get_encrypted()

    def _get_group_name(self):
        """Gets group name from user"""
        if self.group_name:
            group_name = input(
                f"Enter the group name[Leave empty for `{self.group_name}`]:"
            )
            if group_name:
                self.group_name = group_name
        else:
            self.group_name = input("Enter the group name:")

        if "." in self.group_name:
            print("Dots are not allowed in group name!")
            self._get_group_name()
        return self.group_name

    def _get_key_name(self):
        """Gets key name from user"""
        key = input("Enter the name for key:")
        if "." in key:
            print("Dots are not allowed in key name!")
            key = self._get_key_name()
        return key

    def _get_continue_flag(self):
        """Gets continuity flag from user"""
        continue_flag = input("Do you want to add more keys?[Y/N]:").lower()
        if continue_flag == "y":
            return True
        if continue_flag == "n":
            return False
        print("Invalid input!")
        return self._get_continue_flag()

    def add_single_secret(self):
        """Method to add a single secret to the config"""
        print()
        print("*" * 25)
        self.conf.add_new_config(
            group_name=self._get_group_name(),
            key_name=self._get_key_name(),
            unencrypted_values=input("Enter the plain text value:"),
            encrypted=self._get_encrypted(),
            create_group_if_not_exist=True,
            allow_updating=True,
        )

    def save(self, file_name, save_decrypted=False):
        """Saves the configuration"""
        print()
        print("*" * 25)
        if save_decrypted and self.conf.is_config_encrypted:
            print("Decrypting the configuration...")
            self.conf.decrypt_configs()
        if not self.conf.is_config_encrypted and not save_decrypted:
            print("Encrypting the configuration...")
            self.conf.encrypt_configs()

        print("Saving the configuration...")
        self.conf.save(path=f"{file_name}.toml", allow_saving_decrypted=save_decrypted)
        return file_name

    def load_file(self, file_name):
        """loads config file"""
        self.conf.load(f"{file_name}.toml")

    def _loop_for_secret(self):
        """Loops for addition of secrets to the config"""
        while True:
            self.add_single_secret()
            if not self._get_continue_flag():
                return

    def set_encrypted_key(self, encryption_key=None):
        """Abstract method for setting encrypted key"""
        raise NotImplementedError("`set_encrypted_key` not implemented.")

    def interactive(self, file_name, encryption_key=None):
        """Entry method for interactive flow"""
        self.load_file(file_name)
        self.set_encrypted_key(encryption_key)
        self._loop_for_secret()
        self.save(file_name)



"""
Utils related to config
"""

import subprocess
from typing import Any, List

import toml

from archimedes_config.config_manager import ConfigManager
from archimedes_config.utils.path_utils import find_repo_root, resolve_filename_to_base


class RegisteredConfigDecrypted(Exception):
    """Exception to be raised when all the registered config ar enot encrypted"""


def get_secret(file_name, secret_path, *, default_return: Any = KeyError) -> Any:
    """Retrieves secret key from registry"""
    file_name = resolve_filename_to_base(file_name)

    with open(secret_path, "r", encoding="utf8") as file:
        secrets = toml.load(file)
    try:
        return secrets[file_name]
    except KeyError as err:
        if default_return is KeyError:
            raise err
        return default_return


def register_secret(file_name, secret_path, secret_key) -> None:
    """Registers secret key to registry"""
    print("Registering newly generated encryption key...")
    file_name = resolve_filename_to_base(file_name)

    with open(secret_path, "r", encoding="utf8") as file:
        secrets = toml.load(file)
    secrets[file_name] = secret_key

    with open(secret_path, "w", encoding="utf8") as file:
        toml.dump(secrets, file)


def register_config(file_name, config_path) -> None:
    """Registers config to pre commit registry"""
    file_name = resolve_filename_to_base(file_name)
    with open(config_path, "r", encoding="utf8") as file:
        config = toml.load(file)
    if file_name not in config["registered_configs"]:
        config["registered_configs"].append(file_name)
        with open(config_path, "w", encoding="utf8") as file:
            toml.dump(config, file)
    else:
        print("Config already registered.")


def unregister_config(file_name, config_path) -> None:
    """Unregisters config from pre commit registry"""

    with open(config_path, "r", encoding="utf8") as file:
        config = toml.load(file)

    file_name = resolve_filename_to_base(file_name)
    file_name = f"{file_name}.toml"
    if file_name in config["registered_configs"]:
        config["registered_configs"] = [
            i for i in config["registered_configs"] if i != file_name
        ]
    else:
        raise Exception("Provided config was never registered.")
    with open(config_path, "w", encoding="utf8") as file:
        toml.dump(config, file)


def get_registered_configs(config_path) -> List[str]:
    """Returns all the registered configuration sets"""
    with open(config_path, "r", encoding="utf8") as file:
        configs = toml.load(file)
    return configs["registered_configs"]


def check_encrypted(config_path) -> None:
    """Checks if all registered configs are encrypted"""

    root_path = find_repo_root()
    configs = get_registered_configs(config_path)

    committed_files = (
        subprocess.check_output("git diff --name-only --cached".split(" "))
        .decode()
        .split("\n")
    )
    committed_files = [f"{i[:-5]}" for i in committed_files if i.endswith(".toml")]
    committed_files = list(set(configs).intersection(set(committed_files)))
    if len(committed_files) == 0:
        print("Commit doesnt include registered config. Commit is safe.")
        return
    issues = {}
    for config in configs:
        config = root_path / f"{config}.toml"
        try:
            _c = ConfigManager(path=str(config))
        except (FileNotFoundError, IsADirectoryError):
            issues[config] = "FileNotFound"
        else:
            if not _c.get("CONFIGURATIONS", "_IS_CONFIG_ENCRYPTED"):
                issues[config] = "Decrypted"
    if issues:
        issues = "".join(f"\n\t{path} [{issue}]" for path, issue in issues.items())
        issues = f"Issues with config(s) detected: {issues}"
        raise RegisteredConfigDecrypted(issues)
    print("All configurations are encrypted.")



"""
Module for CLI command
"""
import os
import stat
from pathlib import Path

import click
import toml

from archimedes_config.utils.path_utils import clean_filename, find_repo_root
from archimedes_config.utils.util import (
    RegisteredConfigDecrypted,
    check_encrypted,
    get_registered_configs,
    get_secret,
    register_config,
    register_secret,
    unregister_config,
)
from archimedes_config.workflows.add_secrets import AddSecret
from archimedes_config.workflows.lock_secrets import LockConfig
from archimedes_config.workflows.secret_creation import ConfigCreator
from archimedes_config.workflows.unlock_secrets import UnlockConfig

COMMAND_NAME = "arckeyl"
PRE_COMMIT_CHECK_FILE_NAME = f".{COMMAND_NAME}_config"
SECRET_KEYS_FILE_NAME = f".{COMMAND_NAME}_secrets"


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
def cli():
    """
    Welcome to The Archimedes Config Manager.

    \b
    Commands:
        {COMMAND_NAME} new        Create a new config.
        {COMMAND_NAME} add        Add new keys to an existing config.
        {COMMAND_NAME} lock       Locks an existing config.
        {COMMAND_NAME} unlock     Unlocks an existing config.
        {COMMAND_NAME} check      Ensures all registered configs are encrypted.
        {COMMAND_NAME} register      Register config to pre-commit hook.
        {COMMAND_NAME} unregister   Unregister config from pre-commit hook
    """.format(
        COMMAND_NAME=COMMAND_NAME
    )


@cli.command("init")
def init():
    """Initializes pre-commit"""

    root_path = find_repo_root()

    # CREATE CONFIG REGISTRY
    config_path = str(root_path / PRE_COMMIT_CHECK_FILE_NAME)
    if not Path(config_path).exists():
        print("Arckeyl config does not exist. Creating a new config...")
        with open(config_path, "w", encoding="utf8") as file:
            toml.dump({"registered_configs": []}, file)

    # CREATE SECRET REGISTRY
    secrets_path = root_path / SECRET_KEYS_FILE_NAME
    if not secrets_path.exists():
        print("Arckeyl secrets does not exist. Creating a new secrets...")
        with open(secrets_path, "w", encoding="utf8") as file:
            toml.dump({}, file)

    # Add secrets to gitignore
    ignore_path = root_path / ".gitignore"
    if ignore_path.exists():
        with open(ignore_path, "r", encoding="utf8") as file:
            lines = file.readlines()
        if SECRET_KEYS_FILE_NAME not in lines:
            print("Adding secrets to existing gitignore")
            with open(ignore_path, "a", encoding="utf8") as file:
                file.write(f"\n{SECRET_KEYS_FILE_NAME}")
    else:
        print("Created .gitignore and added secrets to it")
        with open(ignore_path, "a", encoding="utf8") as file:
            file.write(f"\n{SECRET_KEYS_FILE_NAME}")

    # CREATE PRE-COMMIT HOOK
    pre_commit_path = root_path / ".git" / "hooks" / "pre-commit"
    command = f"poetry run {COMMAND_NAME} check"
    try:
        with open(pre_commit_path, "r", encoding="utf8") as file:
            lines = file.readlines()
        if command not in lines:
            lines.append(f"\n{command}")
            print("Added encryption check to existing pre-commit webhook")
            with open(pre_commit_path, "w", encoding="utf8") as file:
                file.writelines(lines)
    except FileNotFoundError:
        print("Creating pre commit webhook")
        with open(pre_commit_path, "w", encoding="utf8") as file:
            sample = f"#!/bin/bash\n\n{command}"
            file.write(sample)
    try:
        os.chmod(
            pre_commit_path,
            os.stat(pre_commit_path).st_mode
            | stat.S_IXUSR  # user execute
            | stat.S_IXGRP  # group execute
            | stat.S_IXOTH,  # others execute
        )
    except:  # pylint:disable=bare-except
        print(
            f"Please make sure that the pre commit file located in `{pre_commit_path}` is executable."
        )


@cli.command("add")
@click.argument("file_name")
@clean_filename
def add(file_name):
    """Add a secret to the config"""
    AddSecret().interactive(file_name)


@cli.command("new")
@click.argument("file_name")
@clean_filename
def new(file_name):
    """Create a new config"""
    created_config = ConfigCreator()
    created_config.interactive(file_name)
    repo_root = find_repo_root()
    register_config(file_name, str(repo_root / PRE_COMMIT_CHECK_FILE_NAME))
    if created_config.new_key_generated:
        register_secret(
            file_name,
            str(repo_root / SECRET_KEYS_FILE_NAME),
            created_config.conf.default_key,
        )


@cli.command("lock")
@click.argument("file_name", required=False)
@clean_filename
def lock(file_name):
    """Locks the config"""
    repo_root = find_repo_root()
    if file_name:
        LockConfig().interactive(
            file_name,
            get_secret(
                file_name, str(repo_root / SECRET_KEYS_FILE_NAME), default_return=None
            ),
        )
        return
    config_path = str(repo_root / PRE_COMMIT_CHECK_FILE_NAME)
    configs = get_registered_configs(config_path)
    for config in configs:
        config = str(repo_root / config)
        secret = get_secret(
            config, str(repo_root / SECRET_KEYS_FILE_NAME), default_return=None
        )
        LockConfig().interactive(config, secret)
    print("Locked all configs.")


@cli.command("unlock")
@click.argument("file_name", required=False)
@clean_filename
def unlock(file_name):
    """Unlocks the config"""
    repo_root = find_repo_root()
    if file_name:
        UnlockConfig().interactive(
            file_name, get_secret(file_name, str(repo_root / SECRET_KEYS_FILE_NAME))
        )
        return

    config_path = str(repo_root / PRE_COMMIT_CHECK_FILE_NAME)
    configs = get_registered_configs(config_path)
    for config in configs:
        config = str(repo_root / config)
        secret = get_secret(
            config, str(repo_root / SECRET_KEYS_FILE_NAME), default_return=None
        )
        UnlockConfig().interactive(config, secret)
    print("Unlocked all configs.")


@cli.command("register")
@click.argument("file_name")
@clean_filename
def register(file_name):
    """Adds config file to pre-commit registry"""
    config_path = str(find_repo_root() / PRE_COMMIT_CHECK_FILE_NAME)
    register_config(file_name, config_path)


@cli.command("check")
@clean_filename
def check():
    """Checks encryption status of all registered configs"""
    config_path = str(find_repo_root() / PRE_COMMIT_CHECK_FILE_NAME)
    try:
        check_encrypted(config_path)
    except RegisteredConfigDecrypted as err:
        raise click.ClickException(str(err))


@cli.command("unregister")
@click.argument("file_name")
@clean_filename
def unregister(file_name):
    """Unregister a config from pre-commit registry"""
    config_path = str(find_repo_root() / PRE_COMMIT_CHECK_FILE_NAME)
    unregister_config(file_name, config_path)



"""
Class declaration for configuration manager
"""
from copy import deepcopy
from typing import Any, Dict, Union

import toml
from cryptography.fernet import Fernet


class ConfigManager:
    """
    Manages Configurations
    """

    def __init__(
        self, path: str = None, default_key=None, decrypt_on_load: bool = False
    ) -> None:
        """Init for configuration handler

        Arguments:
            path (str|path): Path to the config file
            default_key (str): Default configuration key to use for encryption / decryption
            decrypt_on_load (bool): Decrypt the config on load
        """
        self._configs = None  # Loaded config
        self.config_loaded_from = (
            None  # Path config loaded from. Used as default saving location.
        )
        self._default_key = None  # Default key

        # if path, set load the config
        if path:
            self.load(path=path)

        # if default key is set, set the key
        if default_key:
            self.set_default_key(default_key)

        if decrypt_on_load and self.is_config_encrypted:
            self.decrypt_configs()

    def __repr__(self) -> str:
        """
        Shows state of configs
        Encrypted configs will not be print encrypted

        :return:
        """
        print_data = {}
        for group, keys in self.config.items():
            print_data[group] = {}
            for key, data in keys.items():
                print_data[group][key] = {}
                for field, value in data.items():
                    print_data[group][key][field] = (
                        "***ENCRYPTED***"
                        if (
                            field == "value"
                            and data.get("encrypted", False)
                            and not self.is_config_encrypted
                        )
                        else value
                    )
        print_data = self.format_toml(print_data)
        return print_data

    @property
    def is_config_encrypted(self) -> bool:
        """Checks the state of encryption if configuration"""
        return self.config["CONFIGURATIONS"]["_IS_CONFIG_ENCRYPTED"]["value"]

    @property
    def default_key(self) -> str:
        """Returns remote key in AZ key vault"""
        if not self._default_key:
            raise ValueError("Default key has not been set.")
        return self._default_key

    @property
    def config(self) -> Dict:
        """Returns configuration"""
        if self._configs:
            return self._configs
        raise ValueError("Configuration has not been loaded.")

    def set_default_key(self, key: str):
        """Sets a default key"""

        if self._default_key:
            raise ValueError("Default key has already been set.")
        self._default_key = key

    def format_toml(self, toml_data: Dict = None):
        """
        Formats input toml into pretty format
        :param toml_data: Dict representation of toml
        :return:
        """
        if toml_data is None:
            toml_data = self.config
        formatted_toml = []
        for group, keys in toml_data.items():
            formatted_toml.append(f"[{group}]")
            for key, value in keys.items():
                formatted_toml.append(f"\t[{group}.{key}]")

                encrypted = value.get("encrypted", False)
                if encrypted:
                    formatted_toml.append("\tencrypted = true")

                value = value["value"]
                if isinstance(value, str):
                    value = f'"{value}"'
                elif isinstance(value, bool):
                    value = "true" if value else "false"
                formatted_toml.append(f"\tvalue = {value}\n")

        return "\n".join(formatted_toml)

    def get(self, group: str, item: str, *, default_return=KeyError) -> Any:
        """
        Returns value for the key
        :param group: Name of group the key belongs to
        :param item: Name of key to extract
        :param default_return: Returns this value if key does not exist in configuration
        :return:
        """
        try:
            key_meta = self.config[group][item]
        except KeyError as err:
            if default_return == KeyError:
                raise err
            return default_return
        if key_meta.get("encrypted", False) and self.is_config_encrypted:
            raise ValueError("Configuration has not been decrypted.")
        return self.config[group][item]["value"]

    def _encrypt_value(self, encrypted: bool, value: Any, key: str = None):
        """
        Encrypts a single value
        :param encrypted: If true, encrypts
        :param value: value to encrypt
        :param key:
        :return:
        """
        key = key if key else self.default_key
        if encrypted:
            value = Fernet(key).encrypt(value.encode()).decode()
        if value is None:
            value = ""
        return value

    def _decrypt_value(self, encrypted: bool, value, key: str = None):
        """
        Decrypts a value
        :param encrypted: If true decrypts
        :param value: Valur to decrypt
        :return:
        """
        key = key if key else self.default_key
        if encrypted:
            value = Fernet(key).decrypt(value).decode()
        if value == "":
            value = None
        return value

    def save(self, path: str = None, allow_saving_decrypted: bool = False) -> None:
        """
        Save configuration into a toml file
        :param path:
            Path to save the configs to
            Defaults to the path the configs were loaded from
        :param allow_saving_decrypted:
            If true, allows saving decrypted configs
        :return:
        """
        if not self.is_config_encrypted:
            if not allow_saving_decrypted:
                raise ValueError(
                    "Saving decrypted configs is not allowed."
                    "Set `allow_saving_decrypted` to True to save decrypted configs."
                )
        if path is None:
            if not self.config_loaded_from:
                raise ValueError("Both default path and specific path not set.")
            path = self.config_loaded_from

        formatted_toml = self.format_toml()
        with open(path, "w", encoding="utf8") as file:
            file.write(formatted_toml)

    def create(self):
        """
        Initializes a new unencrypted configuration
        """
        if self._configs:
            raise ValueError("Configuration has already been loaded.")
        self._configs = {"CONFIGURATIONS": {"_IS_CONFIG_ENCRYPTED": {"value": False}}}

    def load(self, path="config.toml"):
        """
        Loads an environment toml file
        :param path: path to the toml file to read configs from
        :return:
        """
        if self._configs:
            raise ValueError("Configuration has already been loaded.")
        self.config_loaded_from = path
        config = toml.load(path)
        config_copy = deepcopy(config)
        for grp, keys in config.items():
            for field, content in keys.items():
                if "encrypted" in content and not content["encrypted"]:
                    del config_copy[grp][field]["encrypted"]

                if "value" not in content:
                    raise ValueError(
                        f"Invalid toml file. " f"Header {field} doesnt have `value`."
                    )

        self._configs = config_copy

    def list_groups(self):
        """Returns groups"""
        return list(self.config.keys())

    def list_keys(self, group: str):
        """Returns list of keys in configuration"""
        return list(self.config[group].keys())

    def create_new_key(
        self,
        config_save_path: str,
        key_save_path: str,
    ):
        """
        Encrypts an existing toml with a new key

        :param config_save_path: Path to store new config
        :param key_save_path: Path to store new key
        :return:
        """
        new_key = Fernet.generate_key()
        config_encryption_state = self.is_config_encrypted
        if config_encryption_state:
            self.decrypt_configs()
        new_config = self.encrypt_configs(key=new_key.decode(), inplace=False)
        if config_encryption_state:  # revert configuration state to previous
            self.encrypt_configs()

        new_config = self.format_toml(new_config)
        with open(config_save_path, "w", encoding="utf8") as file:
            file.write(new_config)
        with open(key_save_path, "w", encoding="utf8") as file:
            file.write(new_key.decode())

    def encrypt_configs(self, key: str = None, inplace=True) -> Union[Dict, None]:
        """
        Encrypts all the configs
        :param key:
            Key to use for encryption
            Defaults to key set in registry
        :param inplace:
            If true, modifies in memory and returns none
            If false, returns modified data
        :return:
        """
        if self.is_config_encrypted:
            raise ValueError("Config has already been encrypted.")
        key = key if key else self.default_key

        new_encrypted_config = deepcopy(self.config)
        for group, keys in self.config.items():
            for config_key, value in keys.items():
                value = self._encrypt_value(
                    value.get("encrypted", False), value["value"], key
                )
                new_encrypted_config[group][config_key]["value"] = value
        if not inplace:
            return new_encrypted_config
        self._configs = new_encrypted_config
        self._configs["CONFIGURATIONS"]["_IS_CONFIG_ENCRYPTED"]["value"] = True
        return None

    def decrypt_configs(self, bypass_key: str = None, inplace: bool = True):
        """
        Decrypts an env file
        :param bypass_key: If provided, ignores the key in azure registry
        :param inplace:
            If true, modifies in memory and returns none
            If false, returns modified data
        :return:
        """
        if not self.is_config_encrypted:
            raise ValueError("Keys have already been decrypted.")
        enc_key = bypass_key if bypass_key else self.default_key

        decoded_config = deepcopy(self.config)
        for group, keys in self.config.items():
            for key, value in keys.items():
                decoded_config[group][key]["value"] = self._decrypt_value(
                    value.get("encrypted", False), value["value"], enc_key
                )
        if not inplace:
            return decoded_config
        self._configs = decoded_config
        self._configs["CONFIGURATIONS"]["_IS_CONFIG_ENCRYPTED"]["value"] = False
        return None

    def add_new_config(  # pylint:disable=too-many-arguments
        self,
        group_name: str,
        key_name: str,
        unencrypted_values: Any,
        encrypted: bool,
        create_group_if_not_exist: bool = False,
        allow_updating: bool = False,
    ):
        """
        Adds a new configuration

        :param group_name: Group Name that configuration belongs to
        :param key_name: Name for configuration
        :param unencrypted_values: Value to store
        :param encrypted: If true, encrypts the value
        :param create_group_if_not_exist: If false, if group name does not exist raises exception
        :param allow_updating: If false, attempting to update existing key will raise an issue
        :return:
        """
        if "." in group_name:
            raise ValueError("Group name cannot have dot `.`")
        if "." in key_name:
            raise ValueError("Key name cannot have dot `.`")

        if group_name not in self.config:
            if not create_group_if_not_exist:
                raise ValueError(
                    f"Group `{group_name}` doesnt exists."
                    f"Creation of group is not allowed."
                )
            self.config[group_name] = {}

        if not allow_updating:
            if key_name in self.config[group_name]:
                raise ValueError(
                    f"{key_name} already exists in config." f" Updating is not allowed."
                )

        if encrypted and self.is_config_encrypted:
            value = self._encrypt_value(True, unencrypted_values)
        else:
            value = unencrypted_values

        self._configs[group_name][key_name] = {
            "value": value,
            "encrypted": encrypted,
        }



"""
Module for configurations and az vault secrets
"""

import logging

from archimedes_config.config_manager import ConfigManager

try:
    from archimedes_config.keyvault_client import AzureKeyVaultClient
except ImportError as err:
    AzureKeyVaultClient = None


class KeyvaultConfigManager(ConfigManager):  # pylint:disable=too-few-public-methods
    """Class for configurations and az vault secrets"""

    def __init__(
        self, path: str = None, default_key=None, decrypt_on_load: bool = False
    ):
        if AzureKeyVaultClient is None:
            error_message = (
                "Environment has not been setup to handle interactions with Key Vault."
            )
            logging.critical(error_message)
            raise ImportError(error_message)

        super().__init__(
            path=path, default_key=default_key, decrypt_on_load=decrypt_on_load
        )

    def set_default_key_from_key_vault(self):
        """Set default key using secret saved in Azure Key Vault"""

        vault_name = self.config["AZURE_KEYVAULT"]["VAULT_NAME"]["value"]
        key_name = self.config["AZURE_KEYVAULT"]["VAULT_KEY"]["value"]

        self.set_default_key(AzureKeyVaultClient(vault_name).get_secret(name=key_name))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['archimedes_config', 'archimedes_config.utils', 'archimedes_config.workflows']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0', 'cryptography>=39.0.1,<40.0.0', 'toml>=0.10.2,<0.11.0']

extras_require = \
{'azure-keyvault': ['azure-keyvault-secrets', 'azure-identity']}

entry_points = \
{'console_scripts': ['arckeyl = archimedes_config.arckeyl:cli']}

setup_kwargs = {
    'name': 'archimedes-config',
    'version': '0.1.4',
    'description': '',
    'long_description': 'None',
    'author': 'BigyaPradhan',
    'author_email': 'bigya.pradhan@optimeering.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<3.12',
}


setup(**setup_kwargs)



"""
Workflow to add a secret to config
"""

from archimedes_config.workflows.workflow_base import WorkflowBase


class AddSecret(WorkflowBase):
    """Class for adding secret"""

    def set_encrypted_key(self, encryption_key=None):
        if self.config_has_remote_conn_parameters():
            if encryption_key:
                raise Exception(
                    "Config set to connect to azure keyvault. `encryption_key` should not be set."
                )
            self.conf.set_default_key_from_key_vault()
        else:
            self.conf.set_default_key(encryption_key)

        if self.conf.is_config_encrypted:
            self.conf.decrypt_configs()
